/* Program name --- naive.c  2004.9.6 */

#include        <stdio.h>
#include        <string.h>
#include        <stdlib.h>
#include        <math.h>
#include        <time.h>
#include        "multicorr.h"

#define     LOGPRINT            0
#define     RESULT              1

#define     BUFSIZE    100000

/* for memory allocation */
extern void malloc_float();
extern void malloc_int();
extern float *allot_float();
extern int *allot_int();
float *buf_float;
int   *buf_int;
int   float_cnt, int_cnt;
int   limit_float, limit_int;
/* for memory allocation */

void readfile();
void incnaive();
void updateflag();
float incccf();
void error(char *);

main(argc,argv)
int	argc;
char	*argv[];
{
  int   i, j, p, s1, s2;
  int   lag, tick, length, sequences, interval, print_flag;
  char  dbfile[128];
  float *values;

  struct localmaxData buf_locmax[1], *locmax=buf_locmax;
  struct correlationData **corr, *buf_corr;
  struct streamData **stream, *buf_stream;
  struct sequenceData **seq, *buf_seq;

  clock_t timecnt;
  double  stime=0, totaltime;

  if(argc != 5){
    error(argv[0]);
    exit(1);
  }
  strcpy(dbfile,argv[1]);
  length = atoi(argv[2]);
  sequences = atoi(argv[3]);
  interval = atoi(argv[4]);

  /*
   * memory allocation
   */
  corr = (struct correlationData **)malloc(
	 sizeof(struct correlationData *) * sequences * sequences);
  buf_corr = (struct correlationData *)malloc(
         sizeof(struct correlationData) * sequences * sequences);
  seq = (struct sequenceData **)malloc(
	 sizeof(struct sequenceData *) * sequences);
  buf_seq = (struct sequenceData *)malloc(
         sizeof(struct sequenceData) * sequences);
  stream = (struct streamData **)malloc(
         sizeof(struct streamData *) * sequences);
  buf_stream = (struct streamData *)malloc(
         sizeof(struct streamData) * sequences);
  if (corr == NULL || buf_corr == NULL || seq == NULL || buf_seq == NULL || 
      stream == NULL || buf_stream == NULL){
    printf("Allocation error : corr, sqd, stream\n");
    exit(1);
  }

  j = 0;
  /* corr */
  j += (LAGLIMIT+1) * sequences * sequences;
  /* seq */
  j += (LAGLIMIT+1) * sequences * 5;
  /* stream */
  j += length * sequences;
  /* values */
  j += sequences;

  malloc_float(j);

  for (i=0 ; i<sequences*sequences ; i++) {
    corr[i] = buf_corr + i;
    corr[i]->innerprod = allot_float(LAGLIMIT+1);
  }
  for (i=0 ; i<sequences ; i++) {
    seq[i] = buf_seq + i;
    seq[i]->sum_delay    = allot_float(LAGLIMIT+1);
    seq[i]->square_delay = allot_float(LAGLIMIT+1);
    seq[i]->sum_prog     = allot_float(LAGLIMIT+1);
    seq[i]->square_prog  = allot_float(LAGLIMIT+1);
    seq[i]->latest_val  = allot_float(LAGLIMIT+1);
    stream[i] = buf_stream + i;
    stream[i]->value = allot_float(length);
  }
  values = allot_float(sequences);

  /*
   * Read data sequences
   */
  readfile(dbfile, stream, length, sequences);

  /*
   * Initialize sufficient statistics
   */
  for (s1=0 ; s1<sequences ; s1++){
    for (lag=0 ; lag <= LAGLIMIT ; lag++){
      seq[s1]->sum_delay[lag] = 0;
      seq[s1]->square_delay[lag] = 0;
      seq[s1]->sum_prog[lag] = 0;
      seq[s1]->square_prog[lag] = 0;
      for (s2=0 ; s2<sequences ; s2++){
	p = s1 * sequences + s2;
	corr[p]->innerprod[lag] = 0;
      }
    }
  }
  /*
   * Compute lag correlations incrementally and detect them
   */
  for (tick=0 ; tick<length ; tick++){

    /*
     * The flag 'print_flag' means that this program report lag correlations 
     * every 'interval' time-tick. 
     */
    if ((tick+1) % interval == 0){
      print_flag = ON;
#if RESULT
      printf("time-tick %d\n", tick+1);
#endif
    }
    else
      print_flag = OFF;

    /*
     * receive new values at the time-tick 'tick'
     */
    for (s1=0 ; s1<sequences ; s1++)
      values[s1] = stream[s1]->value[tick];

    /*
     * incremental algorithm for naive computation
     */
    timecnt = clock();
    incnaive(seq, corr, locmax, sequences, tick, values, print_flag);
    stime += (double)(clock() - timecnt);

  }
  totaltime = stime / (double)CLOCKS_PER_SEC;
  printf("wall-clock time (Naive) %.4f sec.\n", totaltime);
  printf("\n");

  free(buf_float);
  free(buf_corr);
  free(corr);
  free(buf_stream);
  free(stream);
  free(buf_seq);
  free(seq);
}

void readfile(filename, stream, length, sequences)
char *filename;
struct streamData **stream;
int length;
int sequences;
{

  int i, j;
  char    *strp1,*strp2;
  char  read_buf[BUFSIZE];
  FILE  *fp;

  if ( ( fp = fopen ( filename , "r" ) ) == NULL ){
    printf( "  %s can not open\n" , filename);
    exit(1);
  }

  i = 0;
  while (fgets(read_buf,BUFSIZE,fp) != NULL ){
    strp1 = read_buf;
    j = 0 ;
    while( strp2 = strtok( strp1 , " " )){
      if (strp2 == NULL || strcmp(strp2, "\n") == 0) break;
      stream[j]->value[i] = atof(strp2);
      strp1 = NULL ;
      j++;
      if ( j >= sequences ) break;
    }
    i++;
    if ( i >= length ) break;
  }
  if (i < length)
    printf("Warning: sequence is short. length=%d\n", i);
  fclose(fp);
}

void incnaive(seq, corr, locmax, sequences, tick, values, print_flag)
struct sequenceData **seq;
struct correlationData **corr;
struct localmaxData *locmax;
int sequences;
int tick;
float *values;
int print_flag;
{

  int i, t1;
  int s1, s2, st1, st2;
  int maxlag, m, lag;
  int p, n, nn, flg=OFF;
  float x, r, r0;

  t1 = tick+1;
  maxlag = (int)ceil(t1 / (float)MAXLAG);
  if (maxlag > LAGLIMIT)
    maxlag = LAGLIMIT;
  (tick < LAGLIMIT) ? (m=tick):(m=LAGLIMIT);

  for (s1=0 ; s1<sequences ; s1++){
    /*
     * Pick up the latest value of the stream s1
     */
    x = values[s1];

    /*
     * Compute sufficient statistics (Sx and Sxx)
     *
     * Use sum_delay and square_delay if s1 is delaied by 'lag' time-ticks.
     * Otherwise, use sum_prog and square_prog. 
     */
    for (lag=0 ; lag<=m ; lag++){
      seq[s1]->sum_delay[lag] += x;
      seq[s1]->square_delay[lag] += x*x;
    }
    for (lag=m ; lag>0 ; lag--){
      seq[s1]->sum_prog[lag] = seq[s1]->sum_prog[lag-1];
      seq[s1]->square_prog[lag] = seq[s1]->square_prog[lag-1];
      seq[s1]->latest_val[lag] = seq[s1]->latest_val[lag-1];
    }
    seq[s1]->sum_prog[0] = seq[s1]->sum_delay[0];
    seq[s1]->square_prog[0] = seq[s1]->square_delay[0];
    seq[s1]->latest_val[0] = x;
  }

  /*
   * Compute sufficient statistics (the inner product of x and y, Sxy)
   */
  for (s1=0 ; s1<sequences ; s1++){
    for (s2=0 ; s2<sequences ; s2++){
      if (s1 == s2)
	continue;
      p = s1 * sequences + s2;
      for (lag=0 ; lag<=m ; lag++)
	corr[p]->innerprod[lag] +=
	  seq[s1]->latest_val[0] * seq[s2]->latest_val[lag];
    }
  }
  
  /*
   * Compute lag correlations
   */
  if (print_flag == ON){
    for (s1=0 ; s1<sequences ; s1++){
      for (s2=s1+1 ; s2<sequences ; s2++){
	flg = OFF;
	locmax->num = 0;
	locmax->pd[0].maxr = 0;
	/*
	 * Compute lag correlations for s1 and s2
	 */
	r0 = incccf(seq, corr, s1, s2, sequences, t1, 0);
	updateflag(&flg, locmax, r0, 0);
#if LOGPRINT
printf("0 %f\n", r0);
#endif
	for (lag=1 ; lag<=maxlag ; lag++){
	  r = incccf(seq, corr, s1, s2, sequences, t1, lag);
	  updateflag(&flg, locmax, r, lag);
#if LOGPRINT
printf("%d %f\n", lag, r);
#endif
	}
	if (flg == ON){
	  n = locmax->num;
	  locmax->pd[n-1].end = lag - 1;
	}
	nn = locmax->num;
	flg = OFF;
	locmax->pd[nn].maxr = 0;
	updateflag(&flg, locmax, r0, 0);
#if LOGPRINT
printf("0 %f\n", r0);
#endif
	for (lag=1 ; lag<=maxlag ; lag++){
	  r = incccf(seq, corr, s2, s1, sequences, t1, lag);
	  updateflag(&flg, locmax, r, lag);
#if LOGPRINT
printf("%d %f\n", lag, r);
#endif
	}
	if (flg == ON){
	  n = locmax->num;
	  locmax->pd[n-1].end = lag - 1;
	}
	
	/*
	 * Eliminate the lower local maximum if the correlation coefficient
	 * for lag 0 is larger than the threshold
	 */
	st1 = 0;
	st2 = nn;
	if (fabs(r0) >= THRESHOLD1)
	  (fabs(locmax->pd[0].maxr) <= fabs(locmax->pd[nn].maxr))?
	    (st1++):(st2++);

#if RESULT
	/*
	 * Output lag correlations
	 */
	if (locmax->num > 0){
	  printf("sequences #%d & #%d:\n", s1, s2);
	  printf("#%d delayed by the lag(s) ",s1);
	  for (i=st1 ; i<nn ; i++)
	    printf("%d (%f), ", locmax->pd[i].peak, locmax->pd[i].maxr);
	  printf("\n");
	  printf("#%d delayed by the lag(s) ",s2);
	  for (i=st2 ; i<locmax->num ; i++)
	    printf("%d (%f), ", locmax->pd[i].peak, locmax->pd[i].maxr);
	  printf("\n");
	}
#endif
      }
    }
#if RESULT
    printf("\n");
#endif
  }
}

void updateflag(flg, locmax, y, x)
int *flg;
struct localmaxData *locmax;
float y;
int x;
{
  int n;
  float abs_y;

  n = locmax->num;
  abs_y = fabs(y);

  if (abs_y >= THRESHOLD1){
    if (*flg == OFF){
      if (n<PEAKS){
	locmax->pd[n].maxr = y;
	locmax->pd[n].start = locmax->pd[n].peak = x;
	locmax->num++;
	*flg = ON;
      }
    }
    else{
      if (locmax->pd[n-1].maxr * y < 0){
	locmax->pd[n-1].end = x-1;
	locmax->pd[n].maxr = y;
	locmax->pd[n].start = locmax->pd[n].peak = x;
	locmax->num++;
      }
      else{
	if (fabs(locmax->pd[n-1].maxr) < abs_y){
	  locmax->pd[n-1].maxr = y;
	  locmax->pd[n-1].peak = x;
	}
      }
    }
  }
  else{
    if (*flg == ON){
      if (abs_y < THRESHOLD2 || locmax->pd[n-1].maxr * y < 0){
	locmax->pd[n-1].end = x-1;
	*flg = OFF;
      }
    }
  }
}

float incccf(seq, corr, x, y, sequences, t1, lag)
struct sequenceData **seq;
struct correlationData **corr;
int x;
int y;
int sequences;
int t1;
int lag;
{
  int p;
  float r, c, t, sx, sy, vx, vy;

  p = x * sequences + y;
  t = (float)(t1 - lag);
  sx = seq[x]->sum_delay[lag];
  sy = seq[y]->sum_prog[lag];
  c = corr[p]->innerprod[lag] - sx * sy / t;
  vx = seq[x]->square_delay[lag] - sx * sx / t;
  vy = seq[y]->square_prog[lag] - sy * sy / t;
  r = c / sqrt(vx * vy);

  return r;
}

void error(char *toolname){
  fprintf(stderr,"Usage : %s  dbfile  length  sequences  interval\n",toolname);
  exit(1);
}
